#!/usr/bin/env python
#
# build_icqlink.py - python script to build C++ ICQLink class from 
# icq_Link structure definition in icq.h header file.
#
# Bill Soudan, 3/4/2001
#

# Using the icq_Link struct definition, we generate the following code for 
# the C++ bindings:
#
# 1. virtual methods in ICQLink, one for each callback - these will be
#    called when the callback is invoked, in the appropriate ICQLink
#    instance
# 2. static methods to use as targets for the icq_Link callbacks - both
#    definition in icqlink.h and implementation in icqlink.cpp
# 3. ICQLink constructor code to initialize each callback in icq_Link with 
#    the corresponding static method
#
# We'll use icqlink.h.tpl and icqlink.cpp.tpl as templates.

import string, re

class Callback:

  callback_parse = re.compile(r"^(.+)[(]\*(\w+)[)][(](.*)[)];$");
  arg_parse = re.compile(r"^([\w\s\*]+[\s\*]+)(\w+)|$");

  def __init__(self, definition):
    self.return_type = ""  # callback return type
    self.name = ""         # full callback name
    self.short_name = ""   # callback name with leading 'icq_' removed
    self.arg_list = []     # list of 'type arg_name' strings
    self.arg_name_list = []# list of 'arg_name' strings

    (self.return_type, self.name, args) = self.callback_parse.match(definition).groups()
    self.short_name = self.name[4:]
    for arg in string.split(args, ','):
      arg = string.strip(arg)
      self.arg_list.append(arg)
      (type, name) = self.arg_parse.match(arg).groups()
      self.arg_name_list.append(name)

  def handler_prototype(self):
    return 'handle%s(%s)' % (self.short_name, self.args())

  def virtual_prototype(self):
    return '%s(%s)' % (self.short_name, self.args(1))

  def args(self, start_pos=0):
    return string.join(self.arg_list[start_pos:], ', ')

  def invocation(self):
    return string.join(self.arg_name_list[1:], ', ')

# read icq.h header

print 'Parsing icq.h...'

icq_header = open('/usr/local/include/icq.h', 'r')
raw_lines = icq_header.readlines()
icq_header.close()

# read all callbacks into callback_lines array

in_callbacks = 0;
callback_lines = [];

for line in raw_lines:
  line = string.strip(line)

  if line == '/* End Callbacks */':
    in_callbacks = 0;

  if in_callbacks:
    callback_lines.append(line)

  if line == '/* Begin Callbacks */':
    in_callbacks = 1;

# create callbacks mapping, { callback_name : parameters }

callbacks = {};
callback = ""

for line in callback_lines:
  callback = callback + line;
  if callback[-1] == ';':
    c = Callback(callback)
    callbacks[c.name] = c
    callback = ""

# first, generate the initialization code for icq_Link callbacks

callback_init = []
f = callback_init.append
for callback in callbacks.values():
  f('link->%s = handle%s;' % (callback.name, callback.short_name))

# build implementation for each static handler method

handler_implementation = []
f = handler_implementation.append;
for callback in callbacks.values():
  f('void ICQLink::%s {' % (callback.handler_prototype()))
  f('')
  f('   ICQLink *instance = (ICQLink *)(link);')
  f('   instance->%s(%s); ' % (callback.short_name, callback.invocation()))
  f('')
  f('}')
  f('')

# generate handler method prototypes 

handler_prototypes = []
f = handler_prototypes.append
for callback in callbacks.values():
  f('static void %s;' % callback.handler_prototype())

# generate virtual method prototypes

virtual_prototypes = []
f = virtual_prototypes.append
for callback in callbacks.values():
  f('virtual void %s;' % callback.virtual_prototype())

# finally, create icqlink.h and icqlink.cpp files from templates

subst_re = re.compile(r'\s*(//\*) (\w+) \*//')

def process_template(template, destination):
  template_file = open(template, 'r')
  raw_lines = template_file.readlines()
  template_file.close()

  destination_file = open(destination, 'w')
  destination_file.write('// This file is autogenerated by build_icqlink.py\n')

  for line in raw_lines:
    match = subst_re.match(line)
    if match:
      variable = match.group(2)
      pos = match.start(1)
      for line in globals()[variable]:
        destination_file.write( ' ' * pos + line + '\n')
    else:
      destination_file.write(line)

print 'Generating icqlink.h...'
process_template('icqlink.h.tpl', 'icqlink.h')

print 'Generating icqlink.cpp...'
process_template('icqlink.cpp.tpl', 'icqlink.cpp')

print 'Complete.'

